import numpy as np
import pandas as pd


def run_elo(student_item_scores, k=1):
    """
    Runs the elo algorithm for a given list of students, items, and scores and a weighting parameter k and returns the ability of each student,
     the difficulty of each item, and the Root Mean Squared Error (RMSE) of the predictions.
    :param student_item_scores: a numpy array of shape mxn, where the first column is the student's name, the second column is the item's name and the third
    column is the student's score on that item. A score of 1 is correct and a score of 0 is incorrect.
    :param k: an optional parameter that informs the algorithm how much weight to put on predictions. The range of k is [0, 1], with default value of k 1.
    :return: a dictionary of the abilities of students, a dictionary of the difficulties or items, and the Root Mean Squared Error (RMSE) of predictions.
    """

    if k < 0 or k > 1:
        raise ValueError("k must be between 0 and 1. You provided ", k)

    students = pd.unique(student_item_scores[:, 0])
    items = pd.unique(student_item_scores[:, 1])
    abilities = dict(zip(students, np.zeros(len(students), np.double)))
    difficulties = dict(zip(items, np.zeros(len(items), np.double)))
    predicted = 0.5
    error = 0.0
    for row in student_item_scores:
        correct = int(row[2])
        error += pow(predicted - correct, 2)
        prob = 1 / (1 + np.exp(-(abilities[row[0]] - difficulties[row[1]])))
        abilities[row[0]] = abilities[row[0]] + (k * (correct - prob))
        difficulties[row[1]] = difficulties[row[1]] + (k * (prob - correct))
        predicted = prob
    root_mean_squared_error = np.sqrt(error / len(student_item_scores))
    return abilities, difficulties, root_mean_squared_error



from unittest import TestCase

from numpy.testing import assert_almost_equal

from base_test_case import BaseTestCase
from .elo_runner import run_elo


class Test(TestCase, BaseTestCase):

    def test_run_elo_nine_students_nine_items(self):
        # given
        expected_abilities = {'good-1': 1.1598,
                              'good-2': 0.3651,
                              'good-3': -.1206,
                              'med-1': -0.505,
                              'med-2': -.8437,
                              'med-3': -1.1526,
                              'bad-1': -1.4361,
                              'bad-2': -1.6958,
                              'bad-3': -1.9333}
        expected_difficulties = {'easy-1': -1.1598,
                                 'easy-2': -.3651,
                                 'easy-3': .1206,
                                 'med-1': .5050,
                                 'med-2': .8437,
                                 'med-3': 1.1526,
                                 'hard-1': 1.4361,
                                 'hard-2': 1.6958,
                                 'hard-3': 1.9333}

        # when
        actual_abilities, actual_difficulties, _ = run_elo(self.student_item_scores)

        # then
        for student in actual_abilities.keys():
            assert_almost_equal(actual_abilities[student], expected_abilities[student], decimal=4,
                                err_msg="mismatch:\n", verbose=True)
        for item in actual_difficulties.keys():
            assert_almost_equal(actual_difficulties[item], expected_difficulties[item], decimal=4,
                                err_msg="mismatch:\n", verbose=True)
        assert len(actual_abilities) == len(expected_abilities)
        assert len(actual_difficulties) == len(expected_difficulties)



from unittest import TestCase

from numpy.testing import assert_almost_equal

from .elo_fitter import fit_elo

from base_test_case import BaseTestCase


class Test(TestCase, BaseTestCase):
    def test_fit_elo_nine_students_nine_items(self):
        # given
        expected_optimal_k = 0.99
        expected_lowest_root_mean_squared_error = 0.4899

        # when
        _, _, actual_optimal_k, actual_lowest_root_mean_squared_error = fit_elo(self.student_item_scores)

        # then
        assert_almost_equal(actual_optimal_k, expected_optimal_k, decimal=4, err_msg="mismatch:\n", verbose=True)
        assert_almost_equal(actual_lowest_root_mean_squared_error, expected_lowest_root_mean_squared_error, decimal=4, err_msg="mismatch:\n", verbose=True)



from setuptools import setup, find_packages

VERSION = '0.0.2'
DESCRIPTION = 'Implementation of the ELO Algorithm'
with open('README.md') as f:
    LONG_DESCRIPTION = f.read()

# Setting up
setup(
    name="pcla-elo",
    url='https://pcla.wiki/',
    version=VERSION,
    author="Michael Mogessie",
    author_email="mogessie@upenn.edu",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    install_requires=['numpy', 'pandas'],

    keywords=['python', 'elo'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows"
    ]
)



import sys
from .elo_runner import run_elo


def fit_elo(student_item_scores, step=0.01):
    """
    Searches the optimal k for a given list of students, items, and scores and returns student abilities, item difficulties, the optimal k and
    the Root Mean Squared Error (RMSE) of predictions for that optimal k. This function iteratively calls the run_elo function by varying the values of k
    between 0 and 1, with a default step of 0.01.
    :param student_item_scores: a numpy array of shape mxn, where the first column is the student's name, the second column is the item's name and the third
    column is the student's score on that item. A score of 1 is correct and a score of 0 is incorrect.
    :param step: an optional parameter that informs the algorithm by how much to increase k on each iteration. The default value of step is 0.01.
    :return: a dictionary of the abilities of students, a dictionary of the difficulties or items, the optimal k, and
    the Root Mean Squared Error (RMSE) of predictions.
    """

    if step < 0 or step > 1:
        raise ValueError("step must be between 0 and 1. You provided ", step)

    k = 0
    optimal_k = 0
    lowest_root_mean_squared_error = sys.float_info.max
    optimal_abilities = []
    optimal_difficulties = []
    while k < 1:
        abilities, difficulties, root_mean_squared_error = run_elo(student_item_scores, k)
        if root_mean_squared_error < lowest_root_mean_squared_error:
            optimal_k = k
            lowest_root_mean_squared_error = root_mean_squared_error
            optimal_abilities = abilities
            optimal_difficulties = difficulties
        k += step
    return optimal_abilities, optimal_difficulties, optimal_k, lowest_root_mean_squared_error
